/*********************************************************************/
/*                     Intel Confidential                            */
/*********************************************************************/

/**********************************************************************
**                                                                   **
**                I N T E L   P R O P R I E T A R Y                  **
**                                                                   **
**   COPYRIGHT (c)  2001 - 2011  BY  INTEL  CORPORATION.       ALL   **
**   RIGHTS RESERVED.   NO PART OF THIS PROGRAM OR PUBLICATION MAY   **
**   BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A   **
**   RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER   **
**   LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,   **
**   MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT   **
**   THE PRIOR WRITTEN PERMISSION OF :                               **
**                                                                   **
**                      INTEL  CORPORATION                           **
**                                                                   **
**                2200 MISSION COLLEGE BOULEVARD                     **
**                                                                   **
**             SANTA  CLARA,  CALIFORNIA  95052-8119                 **
**                                                                   **
**********************************************************************/

/**********************************************************************
*                                                                     *
* INTEL CORPORATION PROPRIETARY INFORMATION                           *
* This software is supplied under the terms of a license agreement    *
* with Intel Corporation and may not be copied nor disclosed          *
* except in accordance with the terms of that agreement.              *
*                                                                     *
* Module Name:                                                        *
*   linuxdriver.c                                                     *
*                                                                     *
* Abstract:                                                           *
*   This contains the functions necessary for the Linux Driver        *
*   component of NAL to function as a driver. This includes the       *
*   driver entry point and all Linux specific ring 0 functions.       *
*                                                                     *
***********************************************************************
* VSS Revision Control Information:
* ---------------------------------
*   $Workfile: linuxdriver.c $
*   $Date: 2011/06/20 13:26:27 $
*   $Archive: /QV2.0/nal/src/linux/driver/linuxnaldriver.c $
*   $Revision: 1.3 $
**********************************************************************/

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci_ids.h>
#include <linux/errno.h>
#include <linux/fs.h>           /* Memory mapping functions */
#include <linux/mm.h>           /* Memory mapping functions */
#include <linux/version.h>
#include <asm/uaccess.h>
#include "linuxdriver.h"


static spinlock_t        Global_AtomicTestSetSpinLock;
static char              Global_NalDeviceName[] = NAL_LINUX_DRIVER_NAME;
static int               Global_NalMajor        = 0;

char NalLinuxDriverVersion[] = LINUX_DRIVER_VERSION;

int
NalInitModule(void);

void*
NalAllocIoctlData(
    unsigned int *Size
    );

void
NalFreeIoctlData(
    void *Address
    );

int
NalDeviceControl(
    int  IoctlNumber,
    void *NalIoctlInputData
    );

/* Local function prototypes. */
int
init_module(void);

void
cleanup_module(void);

int
nal_device_control(
    struct inode*   Inode,
    struct file*    File,
    unsigned int    Cmd,
    unsigned long   Arg);

long
nal_device_control_unlocked_ioctl(
    struct file*    File,
    unsigned int    Cmd,
    unsigned long   Arg);

int
nal_open(
    struct inode* Inode,
    struct file*  File);

int
nal_release(
    struct inode*   Inode,
    struct file*    File);


/* This is the file operations structure where all the entry point functions are
 * registered with Linux. */
static struct file_operations Global_NalFops =
{
    owner:          THIS_MODULE,
    /* Starting from kernel version 2.6.36, the 'ioctl' is depreciated.
     * Instead of this we must use 'unlocked_ioctl'.
     * See: http://lwn.net/Articles/400074 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
    ioctl:          nal_device_control,
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
    /* Special include for 2.6.11 and later, where we can call an ioctl
     * without calling a lock on the entire kernel. */
    unlocked_ioctl: nal_device_control_unlocked_ioctl,
#endif
    open:           nal_open,
    release:        nal_release
};


/* These macro calls help setup the modinfo fields in the linux ELF file format,
 * which are then used by modinfo command line utility on linux to provide
 * module information to the user */
MODULE_AUTHOR(DRIVER_COMPANYNAME);
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_LICENSE("Proprietary");
MODULE_VERSION(LINUX_DRIVER_VERSION);

/***************************************************************************
**
** Name:            init_module()
**
** Description:     Called by Linux when the driver is loaded (insmod).
**
** Author:          MVM
**
** Born on Date:    02/19/2003
**
** Arguments:       None
**
** Returns:         on success: 0
**                  on failure: -ENODEV
**
****************************************************************************/
int
init_module(void)
{
    int Result = 0;

    printk(KERN_DEBUG "Intel Pro Diagnostic Driver loading (v. %s)\n", NalLinuxDriverVersion);
    Result = register_chrdev(0, Global_NalDeviceName, &Global_NalFops);
    if(Result < 0)
    {
        Result = -ENODEV;
    }
    else
    {
        Global_NalMajor = Result;
        Result = 0;
    }

    /* Initialize NAL environment */
    NalInitModule();

    /* Init the atomic test/set spinlock */
    spin_lock_init(&Global_AtomicTestSetSpinLock);

    return Result;
}

/***************************************************************************
**
** Name:            cleanup_module()
**
** Description:     Called by Linux when the driver is unloaded (rmmod).
**
** Author:          MVM
**
** Born on Date:    02/19/2003
**
** Arguments:       None
**
** Returns:         None
**
****************************************************************************/
void
cleanup_module(void)
{
    printk(KERN_DEBUG "Intel Pro Diagnostic Driver exiting\n");
    unregister_chrdev(Global_NalMajor, Global_NalDeviceName);
}

/***************************************************************************
**
** Name:            nal_device_control()
**
** Description:     Called by Linux Kernel when an IOCTL is sent to the driver.
**
** Author:          MVM
**
** Born on Date:    02/26/2003
**
** Arguments:       Inode  = Inode file descriptor.
**                  File   = File Descriptor file pointer.
**                  Cmd    = Ioctl Number
**                  Arg    = Ioctl Argument
**
** Returns:         on success: NAL_LINUX_DRIVER_SUCCESS
**                  on failure: NAL_LINUX_DRIVER_FAILED
**
****************************************************************************/
int
nal_device_control(
    struct inode*   Inode,
    struct file*    File,
    unsigned int    Cmd,
    unsigned long   Arg
    )
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    mm_segment_t  old_fs;
#endif
    void          *InputData    = NULL;
    int           Status        = NAL_LINUX_DRIVER_SUCCESS;
    int           InputDataSize = 0;
    int           LeftToCopy    = 0;

    do
    {
        InputData = NalAllocIoctlData(&InputDataSize);
        if(InputData == NULL)
        {
            Status = NAL_LINUX_DRIVER_FAILED;
            break;
        }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        /* The following instructions are used to change the bounds used for
         * checking the validity of pointers in copy_from_user. This expands
         * the useable address space to all user and kernel memory. When we are
         * done we revert back to the old bounds. */
        old_fs = get_fs();
        set_fs(get_ds());
        LeftToCopy = copy_from_user(InputData, (void*)Arg, InputDataSize);
        set_fs(old_fs);
#else
        LeftToCopy = copy_from_user(InputData, (void*)Arg, InputDataSize);
        if(LeftToCopy != 0)
        {
            Status = NAL_LINUX_DRIVER_FAILED;
            break;
        }
#endif

        /* Determine which IOCTL was passed in then call the sub ioctl function
         * for that IOCTL type. */
        Status = NalDeviceControl(Cmd, InputData);
        if(Status != NAL_LINUX_DRIVER_SUCCESS)
        {
            Status = NAL_LINUX_DRIVER_FAILED;
            break;
        }

        /* The following instructions are used to change the bounds used for
         * checking the validity of pointers in copy_to_user. This expands the
         * useable address space to all user and kernel memory. When we are done
         * we revert back to the old bounds. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        old_fs = get_fs();
        set_fs(get_ds());
        LeftToCopy = copy_to_user((void*)Arg, InputData, InputDataSize);
        set_fs(old_fs);
#else

        LeftToCopy = copy_to_user((void*)Arg, InputData, InputDataSize);
        if(LeftToCopy != 0)
        {
            Status = NAL_LINUX_DRIVER_FAILED;
            break;
        }
#endif
    } while(0);

    if(InputData != NULL)
    {
        NalFreeIoctlData(InputData);
    }

    return Status;
}

/***************************************************************************
**
** Name:            nal_device_control_unlocked_ioctl()
**
** Description:     Called by Linux Kernel when an IOCTL is sent to the driver
**                  without locking the kernel.
**
** Author:          OP
**
** Born on Date:    08/20/2007
**
** Arguments:       File   = File Descriptor file pointer.
**                  Cmd    = Ioctl Number
**                  Arg    = Ioctl Argument
**
** Returns:         long
**
****************************************************************************/
long
nal_device_control_unlocked_ioctl(
    struct file*    File,
    unsigned int    Cmd,
    unsigned long   Arg
    )
{
    return nal_device_control(NULL, File, Cmd, Arg);
}

/***************************************************************************
**
** Name:            nal_open()
**
** Description:     Called by Linux Kernel when an Open request is sent to the
**                  driver.
**
** Author:          MVM
**
** Born on Date:    02/26/2003
**
** Arguments:       Inode  = Inode file descriptor.
**                  File   = File Descriptor file pointer.
**
** Returns:         int
**
****************************************************************************/
int
nal_open(
    struct inode* Inode,
    struct file*  File
    )
{
    return 0;
}

/***************************************************************************
**
** Name:            nal_release()
**
** Description:     Called by Linux Kernel when a Release request is sent to
**                  the driver.
**
** Author:          MVM
**
** Born on Date:    02/26/2003
**
** Arguments:       Inode  = Inode file descriptor.
**                  File   = File Descriptor file pointer.
**
** Returns:         int
**
****************************************************************************/
int
nal_release(
    struct inode*   Inode,
    struct file*    File
    )
{
    return 0;
}


/***************************************************************************
**
** Name:            nal_driver_get_version()
**
** Description:     This returns the version number of the driver.
**
** Author:          AS
**
** Born on Date:    03/09/2009
**
** Arguments:       Version - version string
**
** Returns:         None
**
****************************************************************************/
void
nal_driver_get_version(
    char* Version
    )
{
    strcpy(Version, NalLinuxDriverVersion);
}

/***************************************************************************
**
** Name:            nal_spin_lock()
**
** Description:     Spin lock for the module operations.
**
** Author:          PCiunel
**
** Born on Date:    05/19/2011
**
** Arguments:       None
**
** Returns:         None
**
****************************************************************************/
void
nal_spin_lock(void)
{
    spin_lock(&Global_AtomicTestSetSpinLock);
}

/***************************************************************************
**
** Name:            nal_spin_unlock()
**
** Description:     Spin unlock for the module operations.
**
** Author:          PCiunel
**
** Born on Date:    05/19/2011
**
** Arguments:       None
**
** Returns:         None
**
****************************************************************************/
void
nal_spin_unlock(void)
{
    spin_unlock(&Global_AtomicTestSetSpinLock);
}


/***************************************************************************
**
** Name:            nal_get_os_run_domain()
**
** Description:     This returns in which privilege domain kernel runs.
**
** Author:          LKU
**
** Born on Date:    03/24/2011
**
** Arguments:       NONE
**
** Returns:         One of the privilege domains.
**
**
****************************************************************************/
unsigned int
nal_get_os_run_domain(void)
{
    int Domain = NAL_LINUX_DRIVER_OS_DOMAIN_BAREMETAL;

#ifdef is_running_on_xen
    if(is_running_on_xen())
    {
        if(is_initial_xendomain())
        {
            Domain = NAL_LINUX_DRIVER_OS_DOMAIN_0;
        }
        else
        {
            Domain = NAL_LINUX_DRIVER_OS_DOMAIN_U;
        }
    }
#endif

    return Domain;
}

