/*********************************************************************/
/*                     Intel Confidential                            */
/*********************************************************************/

/**********************************************************************
**                                                                   **
**                I N T E L   P R O P R I E T A R Y                  **
**                                                                   **
**   COPYRIGHT (c)  2001 - 2011  BY  INTEL  CORPORATION.       ALL   **
**   RIGHTS RESERVED.   NO PART OF THIS PROGRAM OR PUBLICATION MAY   **
**   BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A   **
**   RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER   **
**   LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,   **
**   MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT   **
**   THE PRIOR WRITTEN PERMISSION OF :                               **
**                                                                   **
**                      INTEL  CORPORATION                           **
**                                                                   **
**                2200 MISSION COLLEGE BOULEVARD                     **
**                                                                   **
**             SANTA  CLARA,  CALIFORNIA  95052-8119                 **
**                                                                   **
**********************************************************************/

/**********************************************************************
**                                                                   **
**  INTEL CORPORATION PROPRIETARY INFORMATION                        **
**                                                                   **
**  This software is supplied under the terms of a license agreement **
**  with Intel Corporation and may not be copied nor disclosed       **
**  except in accordance with the terms of that agreement.           **
**                                                                   **
**                                                                   **
**  Module Name:                                                     **
**    linuxdriverpci_i.c                                             **
**                                                                   **
**  Abstract:                                                        **
**    This file contains function implementations for NAL PCI        **
**    functions for the Linux Driver portion.  These routines        **
**    are called as IOCTL's from the NalResolveIoctl routines.       **
**                                                                   **
***********************************************************************
* VSS Revision Control Information:
* ---------------------------------
*   $Workfile: linuxdriverpci_i.c $
*   $Date: 2011/06/20 13:00:09 $
*   $Archive: /QV2.0/nal/src/linux/driver/linuxdriverpci_i.c $
*   $Revision: 1.16 $
**********************************************************************/
#include <linux/module.h>
#include <linux/pci.h>
#include <asm/byteorder.h>

#include "linuxdriver.h"
#include "linuxdriverpci_i.h"

#ifndef CONFIG_PCI
#error "This driver requires PCI support to be available"
#endif

/***************************************************************************
**
** Name:            nal_pci_get_slot()
**
** Description:     Gets PCI device pointer for the specified PCI slot.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       Segment   = segment number
**                  Bus       = bus number
**                  Device    = device number
**                  Function  = function number
**
** Returns:         Pointer to the PCI device. If the function fails it returns
**                  NULL.
**
****************************************************************************/
void*
nal_pci_get_slot(
    unsigned int Segment,
    unsigned int Bus,
    unsigned int Device,
    unsigned int Function
    )
{
    struct pci_dev*         LinuxPciDevice = NULL;
    struct pci_bus*         LinuxPciBus    = NULL;
    unsigned int            DeviceFunction = 0;

    /* Use segments for creating a Linux PCi Device, for segemented architecture */
    LinuxPciBus = pci_find_bus(Segment, Bus);

    if(LinuxPciBus != NULL)
    {
        /* Set up the DeviceFunction variable */
        DeviceFunction = PCI_DEVFN(Device, Function);

        /* Get the pci device from the bus/dev/function */
        LinuxPciDevice = pci_get_slot(LinuxPciBus, DeviceFunction);
    }
/*    printk(KERN_DEBUG "kernel: pcidev_addr=%d\n", LinuxPciDevice);*/
    return LinuxPciDevice;
}

/***************************************************************************
**
** Name:            nal_pci_enable_device()
**
** Description:     Initialize PCI device (wakes up, assigns its interrupt, ...).
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_enable_device(
    void*          PciDevice
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_enable_device((struct pci_dev*)PciDevice));
}

/***************************************************************************
**
** Name:            nal_pci_disable_device()
**
** Description:     Disable PCI device.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**
** Returns:         None
**
****************************************************************************/
void
nal_pci_disable_device(
    void*               PciDevice
    )
{
    pci_disable_device(PciDevice);
}

/***************************************************************************
**
** Name:            nal_pci_read_config_byte()
**
** Description:     Read one byte from the config space.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Offset    = offset from the beginning of the config space
**                  Value     = read value
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_read_config_byte(
    void*          PciDevice,
    int            Offset,
    unsigned char *Value
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_read_config_byte((struct pci_dev*)PciDevice, Offset, (u8*)Value));
}

/***************************************************************************
**
** Name:            nal_pci_read_config_word()
**
** Description:     Read two bytes from the config space.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Offset    = offset from the beginning of the config space
**                  Value     = read value
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_read_config_word(
    void*          PciDevice,
    int            Offset,
    unsigned short *Value
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_read_config_word((struct pci_dev*)PciDevice, Offset, (u16*)Value));
}

/***************************************************************************
**
** Name:            nal_pci_read_config_dword()
**
** Description:     Read four bytes from the config space.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Offset    = offset from the beginning of the config space
**                  Value     = read value
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_read_config_dword(
    void*          PciDevice,
    int            Offset,
    unsigned int*  Value
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_read_config_dword((struct pci_dev*)PciDevice, Offset, (u32*)Value));
}

/***************************************************************************
**
** Name:            nal_pci_write_config_byte()
**
** Description:     Write one byte to the config space.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Offset    = offset from the beginning of the config space
**                  Value     = write value
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_write_config_byte(
    void*          PciDevice,
    int            Offset,
    unsigned char  Value
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_write_config_byte((struct pci_dev*)PciDevice, Offset, (u8)Value));
}

/***************************************************************************
**
** Name:            nal_pci_write_config_word()
**
** Description:     Write two bytes to the config space.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Offset    = offset from the beginning of the config space
**                  Value     = write value
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_write_config_word(
    void*          PciDevice,
    int            Offset,
    unsigned short Value
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_write_config_word((struct pci_dev*)PciDevice, Offset, (u16)Value));
}

/***************************************************************************
**
** Name:            nal_pci_write_config_dword()
**
** Description:     Write four bytes to the config space.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Offset    = offset from the beginning of the config space
**                  Value     = write value
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_write_config_dword(
    void*          PciDevice,
    int            Offset,
    unsigned int   Value
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_write_config_dword((struct pci_dev*)PciDevice, Offset, (u32)Value));
}

/***************************************************************************
**
** Name:            nal_pci_dev_put()
**
** Description:     Release pointer to the device.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**
** Returns:         Error code.
**
****************************************************************************/
void
nal_pci_dev_put(
    void*          PciDevice
    )
{
    pci_dev_put((struct pci_dev*)PciDevice);
}

/***************************************************************************
**
** Name:            nal_pci_set_dma_mask()
**
** Description:     Set access mask for the DMA read/write operations.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Mask      = mask
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_set_dma_mask(
    void*               PciDevice,
    unsigned long long  Mask
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_set_dma_mask(PciDevice, Mask));
}

/***************************************************************************
**
** Name:            nal_pci_set_consistent_dma_mask()
**
** Description:     Set access mask for the consistent DMA read/write operations.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Mask      = mask
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_set_consistent_dma_mask(
    void*               PciDevice,
    unsigned long long  mask
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_set_consistent_dma_mask(PciDevice, mask));
}

/***************************************************************************
**
** Name:            nal_pci_alloc_consistent()
**
** Description:     Allocation consistent DMA region.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Size      = region size
**                  DmaHandle = handle to the DMA buffer
**
** Returns:         Allocated region.
**
****************************************************************************/
void*
nal_pci_alloc_consistent(
    void*               PciDevice,
    unsigned int        Size,
    unsigned long long* DmaHandle
    )
{
    return pci_alloc_consistent((struct pci_dev*)PciDevice, (size_t)Size, (dma_addr_t*)DmaHandle);
}

/***************************************************************************
**
** Name:            nal_pci_dma_supported()
**
** Description:     Query if PCI controller can properly support DMA addressing.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Mask      = mask
**
** Returns:         Error code.
**
****************************************************************************/
int
nal_pci_dma_supported(
    void*               PciDevice,
    unsigned long long  Mask
    )
{
    return NAL_LINUX_DRIVER_GET_STATUS(pci_dma_supported(PciDevice, Mask));
}

/***************************************************************************
**
** Name:            nal_pci_set_master()
**
** Description:     Enables bus-mastering for pointed device.
**
** Author:          PCiunel
**
** Born on Date:    05/24/2011
**
** Arguments:       PciDevice = device pointer
**                  Mask      = mask
**
** Returns:         None.
**
****************************************************************************/
void
nal_pci_set_master(
    void*               PciDevice
    )
{
    pci_set_master(PciDevice);
}

/***************************************************************************
**
** Name:            nal_pci_resource_start_end()
**
** Description:     Gets first and last address associated with the pointed
**                  region.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       PciDevice = device pointer
**                  Bar       = BAR (base address register) number
**                  Start     = first address to the pointed region
**                  End       = last address of the pointed region
**
** Returns:         Length of the resource region.
**
****************************************************************************/
int
nal_pci_resource_start_end(
    void*                PciDevice,
    unsigned int         Bar,
    unsigned int*        Start,
    unsigned int*        End
    )
{
    *Start = pci_resource_start((struct pci_dev*)PciDevice, Bar);
    *End = pci_resource_end((struct pci_dev*)PciDevice, Bar);
    return pci_resource_len((struct pci_dev*)PciDevice, Bar);
}

/***************************************************************************
**
** Name:            nal_pci_resource_flag()
**
** Description:     Gets flags associated with this resource.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       PciDevice = device pointer
**                  Bar       = BAR (base address register) number
**
** Returns:         Flags.
**
****************************************************************************/
int
nal_pci_resource_flags(
    void*                PciDevice,
    unsigned int         Bar
    )
{
    int FlagsOut = 0;
    int FlagsIn  = pci_resource_flags((struct pci_dev*)PciDevice, Bar);

    NAL_LINUX_KERNEL_GET_RESOURCE_FLAG(FlagsIn, FlagsOut);
    return FlagsOut;
}

/***************************************************************************
**
** Name:            nal_request_region()
**
** Description:     Requests using 'Len' ports starting from 'Start'.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       Start  = first requesting port
**                  Number = number of requesting ports
**                  Name   = device name
**
** Returns:         on success: address to the resources
**                  on failure: NULL
**
****************************************************************************/
void*
nal_request_region(
    unsigned long        Start,
    unsigned long        Number,
    char*                Name
    )
{
    return request_region(Start, Number, Name);
}

/***************************************************************************
**
** Name:            nal_request_mem_region()
**
** Description:     Allocates memory region of 'Len' bytes, starting at 'Start'.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       Start  = first requesting resources
**                  Length = length of requesting region
**                  Name   = device name
**
** Returns:         on success: address to the resources
**                  on failure: NULL
**
****************************************************************************/
void*
nal_request_mem_region(
    unsigned long        Start,
    unsigned long        Length,
    char*                Name
    )
{
    return request_mem_region(Start, Length, Name);
}

/***************************************************************************
**
** Name:            nal_release_region()
**
** Description:     Releases requested region.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       Start  = first requesting port
**                  Number = number of requesting ports
**
** Returns:         None
**
****************************************************************************/
void
nal_release_region(
    unsigned long        Start,
    unsigned long        Len
    )
{
    return release_region(Start, Len);
}

/***************************************************************************
**
** Name:            nal_release_mem_region()
**
** Description:     Releases allocated memory region.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       Start  = first requesting resources
**                  Length = length of requesting region
**
** Returns:         None
**
****************************************************************************/
void
nal_release_mem_region(
    unsigned long        Start,
    unsigned long        Length
    )
{
    return release_mem_region(Start, Length);
}

/***************************************************************************
**
** Name:            nal_release_mem_region()
**
** Description:     Checks availability of memory region.
**
** Author:          PCiunel
**
** Born on Date:    06/06/2011
**
** Arguments:       Start  = first requesting resources
**                  Length = length of requesting region
**
** Returns:         on success: NAL_LINUX_DRIVER_SUCCESS
**                  on failure: NAL_LINUX_DRIVER_FAILD
**
****************************************************************************/
int
nal_check_mem_region(
    unsigned long        Start,
    unsigned long        Length
    )
{
    int Status = check_mem_region(Start, Length);
    return NAL_LINUX_DRIVER_GET_STATUS(Status);
}
